"""
gwalk 是一系列用于管理 Git 仓库的命令行小工具，帮助开发者对大批量的 Git 仓库进行日常维护。
"""

__version__ = "0.2.5"
__status__  = "beta1"
__author__  = "zero <zero.kwok@foxmail.com>"

projectName    = 'gwalk'
projectHome    = 'https://github.com/ZeroKwok/gwalk.git'
projectVersion = f'{__version__} {__status__}'.strip()
projectAuthor  = __author__